"""
####################
# Sequences        #
# (c) 2003 NAUMEN  #
####################

Revision $Id: SequenceManager.py,v 1.2 2003/07/01 09:57:55 void Exp $

Working classes :)

Common idea in implementation is:

Sequence is actually generated by non-persistent singleton with thread-
synchronized methods -- instance of SafeSequence class. Non-persistence protects 
it from multiple instantianting in several ZODB connections.

SequenceStorage acts as persistent interface to SafeSequence -- it creates it, 
when neccessary, initializes with last stored value and remembers last known 
value it passes through. SequenceStorage also resolves ZODB conflicts, that are
caused by it's concurrent use.

SequenceGuard handles SafeSequence's creation and protects from creation race.

SequenceManager just handles several SafeSequences exactly the same way 
SequenceStorage do. It is useful in some designs, but causes more ZODB conflicts 
then several independent SequenceStorages. Conflicts are resolved too, of 
course. """

__version__ = '$Revision: 1.2 $'[11:-2]


import thread, os
from UserDict import UserDict
import ZODB
from Persistence import PersistentMapping
from Globals import DTMLFile
from OFS.SimpleItem import SimpleItem


manage_addSeqStorForm = DTMLFile(os.path.join('dtml','manage_addSeqStorForm'),
                                  globals())

manage_addSeqManForm = DTMLFile(os.path.join('dtml','manage_addSeqManForm'),
                                  globals())

def manage_addSequenceStorage(self, seq_id, RESPONSE):
    """Add Safe Sequence to current folder"""
    self._setObject(seq_id, SequenceStorage(seq_id))
    RESPONSE.redirect(self.DestinationURL() + '/' + seq_id)


def manage_addSequenceManager(self, seqman_id, RESPONSE):
    """Add Safe Sequence to current folder"""
    self._setObject(seqman_id, SequenceManager(seqman_id))
    RESPONSE.redirect(self.DestinationURL() + '/' + seqman_id)



def getSequenceGuard():
    return SG
    # defined in module below



class SequenceStorage(SimpleItem):

    """Peristent interface to SafeSequence.

Creates SafeSequence instance in _getSequence method via call to SequenceGuard, 
if neccessary.
Passes getNextValue calls to this SafeSequence.
Remembers last value returned.
Resolves ZODB conflicts when used simultaneously in several requests.
"""


    meta_type = "Safe Sequence"

    manage_options = ( {'label': 'Contents', 'action' : 'index_html'}, ) \
                     + SimpleItem.manage_options

    index_html = DTMLFile(os.path.join('dtml','sequence'), globals())

    def __init__(self, seq_id, start_value=0):
        self.id = self.seq_id = seq_id
        self.last_value = start_value
        self._getSequence()
        self.title = self.meta_type


    def getNextValue(self):
        """Return next sequence value"""
        seq = self._getSequence()
        result = seq.getNextValue()
        self.last_value = seq.getLastValue()    # may already be not
                                                # equal to 'result'
        return result


    def _getSequence(self):
        if getattr(self, '_v_sequence', None) is None:
            seq = getSequenceGuard().getSequence(self.seq_id, self.last_value)
            self._v_sequence = seq
            self.last_value = seq.getLastValue()
        return self._v_sequence


    def _p_resolveConflict(self, oldState, savedState, newState):
        max_value = max(savedState['last_value'],newState['last_value'])
        savedState['last_value'] = max_value
        return savedState


    def __str__(self):
        return "<Sequence at 0x%08x (SafeSequence at 0x%08x, " \
               "SequenceGuard at 0x%08x)>" % \
               (id(self), id(self._getSequence()), id(getSequenceGuard()))

    def __repr__(self):
        return "SequenceManager.Sequence(%s, %d)" \
                % (self.seq_id, self.last_value)



class SequenceManager(SimpleItem, PersistentMapping):

    """Persistent interface to many SafeSequences

Almost identical to SequenceStorage. Waits for refactoring.
"""

    meta_type = "Safe Sequence Manager"

    index_html = DTMLFile(os.path.join('dtml','seq_man'), globals())

    manage_options = ( {'label': 'Contents', 'action' : 'index_html'}, ) \
                     + SimpleItem.manage_options

    def __init__(self, seqman_id):
        PersistentMapping.__init__(self)
        self.id = seqman_id
        self._v_sequences = {}
        self.title = self.meta_type


    def getNextValue(self, seq_id):
        """Return next value for given sequence"""
        seq = self._getSequence(seq_id)
        result = seq.getNextValue()
        self[seq_id] = seq.getLastValue()
        return result


    def _getSequence(self, seq_id):
        if getattr(self, '_v_sequences', None) is None:
            self._v_sequences = {}
        if self._v_sequences.get(seq_id, None) is None:
            lv =  self.get(seq_id, 0)
            seq = getSequenceGuard().getSequence(seq_id, lv)
            self._v_sequences[seq_id] = seq
            self[seq_id] = seq.getLastValue()
        return self._v_sequences[seq_id]


    def _p_resolveConflict(self, oldState, savedState, newState):
        data = savedState['data']
        new_data = newState['data']
        key_merger = {}
        key_merger.update(data)
        key_merger.update(new_data)
        for seq_id in key_merger.keys():
            sv = data.get(seq_id, 0)
            nv = new_data.get(seq_id, 0)
            data[seq_id] = max(sv, nv)
        return savedState


    def listSequenceIds(self):
        return self.keys()


    def getLastValue(self, seq_id):
        return self[seq_id]


    def __str__(self):
        return "<SequenceManager at 0x%08x (SequenceGuard at 0x%08x)>" % \
               (id(self), id(getSequenceGuard()))



class SequenceGuard(UserDict):

    """Synchronized SafeSequence factory singleton"""

    def __init__(self):
        UserDict.__init__(self)
        self.lock = thread.allocate_lock()


    def getSequence(self, seq_id, init_value):
        self.lock.acquire()
        try:
            if not self.has_key(seq_id):
                self[seq_id] = SafeSequence(init_value)

            seq = self[seq_id]

            if seq.getLastValue() < init_value:
                seq._setLastValue(init_value)

            return seq
        finally:
            self.lock.release()


# Singleton via Python module global
SG = SequenceGuard()


class SafeSequence:

    """Synchronized non-persistent sequence generator. Use as singleton."""

    def __init__(self, init_value):
        self.value = init_value
        self.lock = thread.allocate_lock()


    def getLastValue(self):
        return self.value


    def _setLastValue(self, new_value):
        self.lock.acquire()
        try:
            if new_value > self.value:
                self.value = new_value
        finally:
            self.lock.release()


    def getNextValue(self):
        self.lock.acquire()
        try:
            self.value += 1
            return self.value
        finally:
            self.lock.release()
